@@ -3,29 +3,28 @@
 #
 # furl - URL manipulation made simple.
 #
-# Arthur Grunseid
+# Ansgar Grunseid
 # grunseid.com
 # grunseid@gmail.com
 #
 # License: Build Amazing Things (Unlicense)
+#
+
+from __future__ import division
 
-import sys
 import warnings
 from abc import ABCMeta, abstractmethod
 
 import six
-from six.moves import zip, urllib
+from six.moves import zip
+from six.moves.urllib.parse import (
+    quote, quote_plus, parse_qsl, urlsplit, SplitResult)
 
 import furl
 from furl.omdict1D import omdict1D
-from furl.compat import OrderedDict as odict
+from furl.compat import string_types, OrderedDict as odict
 
-PYTHON_27PLUS = sys.version_info >= (2, 7)
-
-if PYTHON_27PLUS:
-    import unittest
-else:
-    import unittest2 as unittest
+import unittest
 
 
 #
@@ -100,11 +99,11 @@ class itemstr(str, itemcontainer):
         # Keys and values get unquoted. i.e. 'a=a%20a' -> ['a', 'a a']. Empty
         # values without '=' have value None.
         items = []
-        parsed = urllib.parse.parse_qsl(self, keep_blank_values=True)
-        pairstrs = [s2 for s1 in self.split('&')
-                    for s2 in s1.split(';')]
+        parsed = parse_qsl(self, keep_blank_values=True)
+        pairstrs = [
+            s2 for s1 in self.split('&') for s2 in s1.split(';')]
         for (key, value), pairstr in zip(parsed, pairstrs):
-            if key == urllib.parse.quote_plus(pairstr):
+            if key == pairstr:
                 value = None
             items.append((key, value))
         return items
@@ -116,8 +115,28 @@ class itemstr(str, itemcontainer):
         return str(self)
 
 
+class TestMetadata(unittest.TestCase):
+    def test_metadata(self):
+        assert furl.__title__
+        assert furl.__version__
+        assert furl.__license__
+        assert furl.__author__
+        assert furl.__contact__
+        assert furl.__url__
+        assert furl.__description__
+
+
 class TestPath(unittest.TestCase):
 
+    def test_none(self):
+        p = furl.Path(None)
+        assert str(p) == ''
+
+        p = furl.Path('/a/b/c')
+        assert str(p) == '/a/b/c'
+        p.load(None)
+        assert str(p) == ''
+
     def test_isdir_isfile(self):
         for path in ['', '/']:
             p = furl.Path(path)
@@ -189,9 +208,9 @@ class TestPath(unittest.TestCase):
         for path in encoded:
             assert str(furl.Path(path)) == path
 
+        safe = furl.Path.SAFE_SEGMENT_CHARS + '/'
         for path in decoded:
-            assert str(furl.Path(path)) == urllib.parse.quote(
-                path, "/:@-._~!$&'()*+,;=")
+            assert str(furl.Path(path)) == quote(path, safe)
 
         # Valid path segment characters should not be encoded.
         for char in ":@-._~!$&'()*+,;=":
@@ -202,7 +221,7 @@ class TestPath(unittest.TestCase):
         # Invalid path segment characters should be encoded.
         for char in ' ^`<>[]"?':
             f = furl.furl().set(path=char)
-            assert str(f.path) == f.url == urllib.parse.quote(char)
+            assert str(f.path) == f.url == quote(char)
             assert f.path.segments == [char]
 
         # Encode '/' within a path segment.
@@ -212,6 +231,13 @@ class TestPath(unittest.TestCase):
         assert f.path.segments == [segment]
         assert f.url == 'a%2Fb'
 
+        # Encode percent signs in path segment stings.
+        assert str(furl.Path(['a%20d'])) == 'a%2520d'
+        assert str(furl.Path(['a%zzd'])) == 'a%25zzd'
+
+        # Percent-encodings should be capitalized, as per RFC 3986.
+        assert str(furl.Path('a%2fd')) == str(furl.Path('a%2Fd')) == 'a%2Fd'
+
     def test_load(self):
         self._test_set_load(furl.Path.load)
 
@@ -220,15 +246,17 @@ class TestPath(unittest.TestCase):
 
     def _test_set_load(self, path_set_or_load):
         p = furl.Path('a/b/c/')
+        assert path_set_or_load(p, furl.Path('asdf/asdf/')) == p
+        assert not p.isabsolute and str(p) == 'asdf/asdf/'
+
         assert path_set_or_load(p, 'asdf/asdf/') == p
-        assert not p.isabsolute
-        assert str(p) == 'asdf/asdf/'
+        assert not p.isabsolute and str(p) == 'asdf/asdf/'
+
         assert path_set_or_load(p, ['a', 'b', 'c', '']) == p
-        assert not p.isabsolute
-        assert str(p) == 'a/b/c/'
+        assert not p.isabsolute and str(p) == 'a/b/c/'
+
         assert path_set_or_load(p, ['', 'a', 'b', 'c', '']) == p
-        assert p.isabsolute
-        assert str(p) == '/a/b/c/'
+        assert p.isabsolute and str(p) == '/a/b/c/'
 
     def test_add(self):
         # URL paths.
@@ -457,10 +485,10 @@ class TestPath(unittest.TestCase):
             assert p.normalize() is p and str(p) == str(p.normalize()) == path
 
         # Path modified.
-        tonormalize = [
+        to_normalize = [
             ('//', '/'), ('//a', '/a'), ('//a/', '/a/'), ('//a///', '/a/'),
             ('////a/..//b', '/b'), ('/a/..//b//./', '/b/')]
-        for path, normalized in tonormalize:
+        for path, normalized in to_normalize:
             p = furl.Path(path)
             assert p.normalize() is p and str(p.normalize()) == normalized
 
@@ -499,6 +527,58 @@ class TestPath(unittest.TestCase):
             p = furl.Path(path)
             assert str(p) == path_encoded
 
+    def test_itruediv(self):
+        p = furl.Path()
+
+        p /= 'a'
+        assert str(p) == 'a'
+
+        p /= 'b'
+        assert str(p) == 'a/b'
+
+        p /= 'c d/'
+        assert str(p) == 'a/b/c%20d/'
+
+        p /= furl.Path('e')
+        assert str(p) == 'a/b/c%20d/e'
+
+    def test_truediv(self):
+        p = furl.Path()
+
+        p1 = p / 'a'
+        assert p1 is not p
+        assert str(p1) == 'a'
+
+        p2 = p / 'a' / 'b'
+        assert p2 is not p
+        assert str(p) == ''
+        assert str(p2) == 'a/b'
+
+        # Path objects should be joinable with other Path objects.
+        p3 = furl.Path('e')
+        p4 = furl.Path('f')
+        assert p3 / p4 == furl.Path('e/f')
+
+        # Joining paths with __truediv__ should not modify the original, even
+        # if <isabsolute> is True.
+        p5 = furl.Path(['a', 'b'], force_absolute=lambda _: True)
+        p6 = p5 / 'c'
+        assert str(p5) == '/a/b'
+        assert str(p6) == '/a/b/c'
+
+    def test_asdict(self):
+        segments = ['wiki', 'ロリポップ']
+        path_encoded = 'wiki/%E3%83%AD%E3%83%AA%E3%83%9D%E3%83%83%E3%83%97'
+        p = furl.Path(path_encoded)
+        d = {
+            'isdir': False,
+            'isfile': True,
+            'isabsolute': False,
+            'segments': segments,
+            'encoded': path_encoded,
+            }
+        assert p.asdict() == d
+
 
 class TestQuery(unittest.TestCase):
 
@@ -543,7 +623,7 @@ class TestQuery(unittest.TestCase):
             'a=a&c=c%5Ec', '<=>&^="', '%3C=%3E&%5E=%22', '%=%;`=`',
             '%25=%25&%60=%60',
             # Only keys, no values.
-            'asdfasdf', '/asdf/asdf/sdf', '*******', '!@#(*&@!#(*@!#', 'a&b&',
+            'asdfasdf', '/asdf/asdf/sdf', '*******', '!@#(*&@!#(*@!#', 'a&b',
             'a;b',
             # Repeated parameters.
             'a=a&a=a', 'space=a+a&space=b+b',
@@ -555,11 +635,19 @@ class TestQuery(unittest.TestCase):
         self.items = (self.itemlists + self.itemdicts + self.itemomdicts +
                       self.itemstrs)
 
+    def test_none(self):
+        q = furl.Query(None)
+        assert str(q) == ''
+
+        q = furl.Query('a=b&c=d')
+        assert str(q) == 'a=b&c=d'
+        q.load(None)
+        assert str(q) == ''
+
     def test_various(self):
         for items in self.items:
             q = furl.Query(items.original())
             assert q.params.allitems() == items.allitems()
-            pairs = ['%s=%s' % (i[0], i[1]) for i in self._quote_items(items)]
 
             # encode() accepts both 'delimiter' and 'delimeter'. The
             # latter was incorrectly used until furl v0.4.6.
@@ -764,14 +852,70 @@ class TestQuery(unittest.TestCase):
         assert query.encode(';') == 'a=b+c;d=e+f'
         assert query.encode(';', quote_plus=False) == 'a=b%20c;d=e%20f'
 
+        # Encode '/' consistently across quote_plus=True and quote_plus=False.
+        query = furl.Query('a /b')
+        assert query.encode(quote_plus=True) == 'a+%2Fb'
+        assert query.encode(quote_plus=False) == 'a%20%2Fb'
+
+        # dont_quote= accepts both True and a string of safe characters not to
+        # percent-encode. Unsafe query characters, like '^' and '#', are always
+        # percent-encoded.
+        query = furl.Query('a %2B/b?#')
+        assert query.encode(dont_quote='^') == 'a+%2B%2Fb%3F%23'
+        assert query.encode(quote_plus=True, dont_quote=True) == 'a++/b?%23'
+        assert query.encode(quote_plus=False, dont_quote=True) == 'a%20+/b?%23'
+
+    def test_asdict(self):
+        pairs = [('a', '1'), ('ロリポップ', 'testä')]
+        key_encoded = '%E3%83%AD%E3%83%AA%E3%83%9D%E3%83%83%E3%83%97'
+        value_encoded = 'test%C3%A4'
+        query_encoded = 'a=1&' + key_encoded + '=' + value_encoded
+        p = furl.Query(query_encoded)
+        d = {
+            'params': pairs,
+            'encoded': query_encoded,
+            }
+        assert p.asdict() == d
+
+    def test_value_encoding_empty_vs_nonempty_key(self):
+        pair = ('=', '=')
+        pair_encoded = '%3D=%3D'
+        assert furl.Query(pair_encoded).params.allitems() == [pair]
+
+        q = furl.Query()
+        q.params = [pair]
+        assert q.encode() == pair_encoded
+
+        empty_key_pair = ('', '==3===')
+        empty_key_encoded = '===3==='
+        assert furl.Query(empty_key_encoded).params.items() == [empty_key_pair]
+
+    def test_special_characters(self):
+        q = furl.Query('==3==')
+        assert q.params.allitems() == [('', '=3==')] and str(q) == '==3=='
+
+        f = furl.furl('https://www.google.com????')
+        assert f.args.allitems() == [('???', None)]
+
+        q = furl.Query('&=&')
+        assert q.params.allitems() == [('', None), ('', ''), ('', None)]
+        assert str(q) == '&=&'
+
+        url = 'https://www.google.com?&&%3F=&%3F'
+        f = furl.furl(url)
+        assert f.args.allitems() == [
+            ('', None), ('', None), ('?', ''), ('?', None)]
+        assert f.url == url
+
     def _quote_items(self, items):
         # Calculate the expected querystring with proper query encoding.
         #   Valid query key characters: "/?:@-._~!$'()*,;"
         #   Valid query value characters: "/?:@-._~!$'()*,;="
         allitems_quoted = []
         for key, value in items.iterallitems():
-            pair = (urllib.parse.quote_plus(str(key), "/?:@-._~!$'()*,;"),
-                    urllib.parse.quote_plus(str(value), "/?:@-._~!$'()*,;="))
+            pair = (
+                quote_plus(str(key), "/?:@-._~!$'()*,;"),
+                quote_plus(str(value), "/?:@-._~!$'()*,;="))
             allitems_quoted.append(pair)
         return allitems_quoted
 
@@ -793,6 +937,9 @@ class TestQueryCompositionInterface(unittest.TestCase):
         assert isinstance(t.query, furl.Query)
         assert str(t.query) == ''
 
+        t.args = {'55': '66'}
+        assert t.args == {'55': '66'} and str(t.query) == '55=66'
+
         t.query = 'a=a&s=s s'
         assert isinstance(t.query, furl.Query)
         assert str(t.query) == 'a=a&s=s+s'
@@ -822,6 +969,15 @@ class TestFragment(unittest.TestCase):
         f.query.params['hi'] = 'gr8job'
         assert str(f) == 'no=dads&hi=gr8job'
 
+    def test_none(self):
+        f = furl.Fragment(None)
+        assert str(f) == ''
+
+        f = furl.Fragment('sup')
+        assert str(f) == 'sup'
+        f.load(None)
+        assert str(f) == ''
+
     def test_load(self):
         comps = [('', '', {}),
                  ('?', '%3F', {}),
@@ -838,13 +994,13 @@ class TestFragment(unittest.TestCase):
                  ('dog?machine?yes', 'dog%3Fmachine%3Fyes', {}),
                  ('dog?machine=?yes', 'dog', {'machine': '?yes'}),
                  ('schtoot?a=a&hok%20sprm', 'schtoot',
-                  {'a': 'a', 'hok sprm': ''}),
+                  {'a': 'a', 'hok sprm': None}),
                  ('schtoot?a=a&hok sprm', 'schtoot',
-                  {'a': 'a', 'hok sprm': ''}),
+                  {'a': 'a', 'hok sprm': None}),
                  ('sch/toot?a=a&hok sprm', 'sch/toot',
-                  {'a': 'a', 'hok sprm': ''}),
+                  {'a': 'a', 'hok sprm': None}),
                  ('/sch/toot?a=a&hok sprm', '/sch/toot',
-                  {'a': 'a', 'hok sprm': ''}),
+                  {'a': 'a', 'hok sprm': None}),
                  ]
 
         for fragment, path, query in comps:
@@ -952,6 +1108,25 @@ class TestFragment(unittest.TestCase):
         f.separator = False
         assert f
 
+    def test_asdict(self):
+        path_encoded = '/wiki/%E3%83%AD%E3%83%AA%E3%83%9D%E3%83%83%E3%83%97'
+
+        key_encoded = '%E3%83%AD%E3%83%AA%E3%83%9D%E3%83%83%E3%83%97'
+        value_encoded = 'test%C3%A4'
+        query_encoded = 'a=1&' + key_encoded + '=' + value_encoded
+
+        fragment_encoded = path_encoded + '?' + query_encoded
+        p = furl.Path(path_encoded)
+        q = furl.Query(query_encoded)
+        f = furl.Fragment(fragment_encoded)
+        d = {
+            'separator': True,
+            'path': p.asdict(),
+            'query': q.asdict(),
+            'encoded': fragment_encoded,
+            }
+        assert f.asdict() == d
+
 
 class TestFragmentCompositionInterface(unittest.TestCase):
 
@@ -1007,9 +1182,24 @@ class TestFurl(unittest.TestCase):
         # So, as a result of using urlparse.urlsplit(), this little helper
         # function only works when provided URLs whos schemes are also in
         # urlparse.uses_query.
-        items = urllib.parse.parse_qsl(urllib.parse.urlsplit(url).query, True)
+        items = parse_qsl(urlsplit(url).query, True)
         return (key, val) in items
 
+    def test_constructor_and_set(self):
+        f = furl.furl(
+            'http://user:pass@pumps.ru/', args={'hi': 'bye'},
+            scheme='scrip', path='prorp', host='horp', fragment='fraggg')
+        assert f.url == 'scrip://user:pass@horp/prorp?hi=bye#fraggg'
+
+    def test_none(self):
+        f = furl.furl(None)
+        assert str(f) == ''
+
+        f = furl.furl('http://user:pass@pumps.ru/')
+        assert str(f) == 'http://user:pass@pumps.ru/'
+        f.load(None)
+        assert str(f) == ''
+
     def test_idna(self):
         decoded_host = u'ドメイン.テスト'
         encoded_url = 'http://user:pass@xn--eckwd4c7c.xn--zckzah/'
@@ -1026,7 +1216,7 @@ class TestFurl(unittest.TestCase):
         assert f.url == encoded_url
 
         f = furl.furl().set(host=u'ロリポップ')
-        assert f.url == 'xn--9ckxbq5co'
+        assert f.url == '//xn--9ckxbq5co'
 
     def test_unicode(self):
         paths = ['ロリポップ', u'ロリポップ']
@@ -1062,8 +1252,8 @@ class TestFurl(unittest.TestCase):
             assert f.args[key] == value  # Unicode values aren't modified.
             assert key not in f.url
             assert value not in f.url
-            assert urllib.parse.quote_plus(furl.utf8(key)) in f.url
-            assert urllib.parse.quote_plus(furl.utf8(value)) in f.url
+            assert quote_plus(furl.utf8(key)) in f.url
+            assert quote_plus(furl.utf8(value)) in f.url
             f.path.segments = [path]
             assert f.path.segments == [path]  # Unicode values aren't modified.
             assert f.url == full_url_encoded
@@ -1083,33 +1273,39 @@ class TestFurl(unittest.TestCase):
         # Protocol relative URLs.
         for url in ['//', '//sup.txt', '//arc.io/d/sup']:
             f = furl.furl(url)
-            assert f.scheme == '' and f.url == url
+            assert f.scheme is None and f.url == url
 
         f = furl.furl('//sup.txt')
-        assert f.scheme == ''
-        f.scheme = None
-        assert f.scheme is None and f.url == 'sup.txt'
+        assert f.scheme is None and f.url == '//sup.txt'
         f.scheme = ''
-        assert f.scheme == '' and f.url == '//sup.txt'
+        assert f.scheme == '' and f.url == '://sup.txt'
 
-        # Schemes without slashes , like 'mailto:'.
-        f = furl.furl('mailto:sup@sprp.ru')
-        assert f.url == 'mailto:sup@sprp.ru'
-        f = furl.furl('mailto://sup@sprp.ru')
-        assert f.url == 'mailto:sup@sprp.ru'
+        # Schemes without slashes, like 'mailto:'.
+        assert furl.furl('mailto:sup@sprp.ru').url == 'mailto:sup@sprp.ru'
+        assert furl.furl('mailto://sup@sprp.ru').url == 'mailto://sup@sprp.ru'
 
         f = furl.furl('mailto:sproop:spraps@sprp.ru')
         assert f.scheme == 'mailto'
-        assert f.username == 'sproop' and f.password == 'spraps'
-        assert f.host == 'sprp.ru'
+        assert f.path == 'sproop:spraps@sprp.ru'
 
         f = furl.furl('mailto:')
-        assert f.url == 'mailto:' and f.scheme == 'mailto'
+        assert f.url == 'mailto:' and f.scheme == 'mailto' and f.netloc is None
+
+        f = furl.furl('tel:+1-555-555-1234')
+        assert f.scheme == 'tel' and str(f.path) == '+1-555-555-1234'
+
+        f = furl.furl('urn:srp.com/ferret?query')
+        assert f.scheme == 'urn' and str(f.path) == 'srp.com/ferret'
+        assert str(f.query) == 'query'
 
         # Ignore invalid schemes.
         assert furl.furl('+invalid$scheme://lolsup').scheme is None
         assert furl.furl('/api/test?url=http://a.com').scheme is None
 
+        # Empty scheme.
+        f = furl.furl(':')
+        assert f.scheme == '' and f.netloc is None and f.url == ':'
+
     def test_username_and_password(self):
         # Empty usernames and passwords.
         for url in ['', 'http://www.pumps.com/']:
@@ -1122,7 +1318,7 @@ class TestFurl(unittest.TestCase):
 
         # Username only.
         for username in usernames:
-            encoded_username = urllib.parse.quote(username, safe='')
+            encoded_username = quote(username, safe='')
             encoded_url = 'http://%s@www.google.com/' % encoded_username
 
             f = furl.furl(encoded_url)
@@ -1143,7 +1339,7 @@ class TestFurl(unittest.TestCase):
 
         # Password only.
         for password in passwords:
-            encoded_password = urllib.parse.quote(password, safe='')
+            encoded_password = quote(password, safe='')
             encoded_url = 'http://:%s@www.google.com/' % encoded_password
 
             f = furl.furl(encoded_url)
@@ -1165,8 +1361,8 @@ class TestFurl(unittest.TestCase):
         # Username and password.
         for username in usernames:
             for password in passwords:
-                encoded_username = urllib.parse.quote(username, safe='')
-                encoded_password = urllib.parse.quote(password, safe='')
+                encoded_username = quote(username, safe='')
+                encoded_password = quote(password, safe='')
                 encoded_url = 'http://%s:%s@www.google.com/' % (
                     encoded_username, encoded_password)
 
@@ -1206,19 +1402,27 @@ class TestFurl(unittest.TestCase):
         f = furl.furl()
         assert f.username is f.password is None
         f.username = 'uu'
-        assert f.username == 'uu' and f.password is None and f.url == 'uu@'
+        assert f.username == 'uu' and f.password is None and f.url == '//uu@'
         f.password = 'pp'
-        assert f.username == 'uu' and f.password == 'pp' and f.url == 'uu:pp@'
+        assert f.username == 'uu' and f.password == 'pp'
+        assert f.url == '//uu:pp@'
         f.username = ''
-        assert f.username == '' and f.password == 'pp' and f.url == ':pp@'
+        assert f.username == '' and f.password == 'pp' and f.url == '//:pp@'
         f.password = ''
-        assert f.username == f.password == '' and f.url == ':@'
+        assert f.username == f.password == '' and f.url == '//:@'
         f.password = None
-        assert f.username == '' and f.password is None and f.url == '@'
+        assert f.username == '' and f.password is None and f.url == '//@'
         f.username = None
         assert f.username is f.password is None and f.url == ''
         f.password = ''
-        assert f.username is None and f.password == '' and f.url == ':@'
+        assert f.username is None and f.password == '' and f.url == '//:@'
+
+        # Unicode.
+        username = u'kødp'
+        password = u'ålæg'
+        f = furl.furl(u'https://%s:%s@example.com/' % (username, password))
+        assert f.username == username and f.password == password
+        assert f.url == 'https://k%C3%B8dp:%C3%A5l%C3%A6g@example.com/'
 
     def test_basics(self):
         url = 'hTtP://www.pumps.com/'
@@ -1284,10 +1488,21 @@ class TestFurl(unittest.TestCase):
         # with '//', as-is the default behavior of
         # urlparse.urlunsplit().
         f = furl.furl()
-        assert f.set(host='foo').url == 'foo'
-        assert f.set(host='pumps.com').url == 'pumps.com'
-        assert f.set(host='pumps.com', port=88).url == 'pumps.com:88'
-        assert f.set(netloc='pumps.com:88').url == 'pumps.com:88'
+        assert f.set(host='foo').url == '//foo'
+        assert f.set(host='pumps.com').url == '//pumps.com'
+        assert f.set(host='pumps.com', port=88).url == '//pumps.com:88'
+        assert f.set(netloc='pumps.com:88').url == '//pumps.com:88'
+
+        # furl('...') and furl.url = '...' are functionally identical.
+        url = 'https://www.pumps.com/path?query#frag'
+        f1 = furl.furl(url)
+        f2 = furl.furl()
+        f2.url = url
+        assert f1 == f2
+
+        # Empty scheme and netloc.
+        f = furl.furl('://')
+        assert f.scheme == f.netloc == '' and f.url == '://'
 
     def test_basic_manipulation(self):
         f = furl.furl('http://www.pumps.com/')
@@ -1329,6 +1544,31 @@ class TestFurl(unittest.TestCase):
         f.host = 'ohay.com'
         assert str(f) == 'sup://ohay.com/hay%20supppp?space=1+2#sup'
 
+    def test_path_itruediv(self):
+        f = furl.furl('http://www.pumps.com/')
+
+        f /= 'a'
+        assert f.url == 'http://www.pumps.com/a'
+
+        f /= 'b'
+        assert f.url == 'http://www.pumps.com/a/b'
+
+        f /= 'c d/'
+        assert f.url == 'http://www.pumps.com/a/b/c%20d/'
+
+    def test_path_truediv(self):
+        f = furl.furl('http://www.pumps.com/')
+
+        f1 = f / 'a'
+        assert f.url == 'http://www.pumps.com/'
+        assert f1.url == 'http://www.pumps.com/a'
+
+        f2 = f / 'c' / 'd e/'
+        assert f2.url == 'http://www.pumps.com/c/d%20e/'
+
+        f3 = f / furl.Path('f')
+        assert f3.url == 'http://www.pumps.com/f'
+
     def test_odd_urls(self):
         # Empty.
         f = furl.furl('')
@@ -1347,8 +1587,13 @@ class TestFurl(unittest.TestCase):
             "sup://example.com/:@-._~!$&'()*+,=;:@-._~!$&'()*+,=:@-._~!$&'()*+"
             ",==?/?:@-._~!$'()*+,;=/?:@-._~!$'()*+,;==#/?:@-._~!$&'()*+,;=")
         pathstr = "/:@-._~!$&'()*+,=;:@-._~!$&'()*+,=:@-._~!$&'()*+,=="
-        querystr = "/?:@-._~!$'()*+,=&=/?:@-._~!$'()*+,&=="
-        fragmentstr = "/?:@-._~!$=&'()*+,=&="
+        querystr = (
+            quote_plus("/?:@-._~!$'()* ,") + '&' +
+            '=' + quote_plus("/?:@-._~!$'()* ,") + '&' +
+            '==')
+        fragmentstr = (
+            '/?' + quote_plus(':@-._~!$') + '&' +
+            quote_plus("'()* ,") + '&' + '=')
         f = furl.furl(url)
         assert f.scheme == 'sup'
         assert f.host == 'example.com'
@@ -1361,36 +1606,47 @@ class TestFurl(unittest.TestCase):
         # Scheme only.
         f = furl.furl('sup://')
         assert f.scheme == 'sup'
-        assert f.host is f.port is f.netloc is None
-        assert str(f.path) == ''
-        assert str(f.query) == ''
+        assert f.host == f.netloc == ''
+        assert f.port is None
+        assert str(f.path) == str(f.query) == str(f.fragment) == ''
         assert f.args == f.query.params == {}
-        assert str(f.fragment) == ''
-        assert f.url == 'sup://' and f.netloc is None
+        assert f.url == 'sup://'
         f.scheme = None
-        assert f.scheme is None and f.netloc is None and f.url == ''
+        assert f.scheme is None and f.netloc == '' and f.url == '//'
         f.scheme = ''
-        assert f.scheme == '' and f.netloc is None and f.url == '//'
+        assert f.scheme == '' and f.netloc == '' and f.url == '://'
+
+        f = furl.furl('sup:')
+        assert f.scheme == 'sup'
+        assert f.host is f.port is f.netloc is None
+        assert str(f.path) == str(f.query) == str(f.fragment) == ''
+        assert f.args == f.query.params == {}
+        assert f.url == 'sup:'
+        f.scheme = None
+        assert f.url == '' and f.netloc is None
+        f.scheme = ''
+        assert f.url == ':' and f.netloc is None
 
         # Host only.
         f = furl.furl().set(host='pumps.meat')
-        assert f.url == 'pumps.meat' and f.netloc == f.host == 'pumps.meat'
+        assert f.url == '//pumps.meat' and f.netloc == f.host == 'pumps.meat'
         f.host = None
         assert f.url == '' and f.host is f.netloc is None
         f.host = ''
-        assert f.url == '' and f.host == f.netloc == ''
+        assert f.url == '//' and f.host == f.netloc == ''
 
         # Port only.
         f = furl.furl()
         f.port = 99
-        assert f.url == ':99' and f.netloc is not None
+        assert f.url == '//:99' and f.netloc is not None
         f.port = None
         assert f.url == '' and f.netloc is None
 
         # urlparse.urlsplit() treats the first two '//' as the beginning
         # of a netloc, even if the netloc is empty.
         f = furl.furl('////path')
-        assert f.url == '//path' and str(f.path) == '//path'
+        assert f.netloc == '' and str(f.path) == '//path'
+        assert f.url == '////path'
 
         # TODO(grun): Test more odd URLs.
 
@@ -1398,7 +1654,7 @@ class TestFurl(unittest.TestCase):
         # No host.
         url = 'http:///index.html'
         f = furl.furl(url)
-        assert f.host is None and furl.furl(url).url == url
+        assert f.host == '' and furl.furl(url).url == url
 
         # Valid IPv4 and IPv6 addresses.
         f = furl.furl('http://192.168.1.101')
@@ -1422,15 +1678,12 @@ class TestFurl(unittest.TestCase):
         # exception on invalid IPv6 addresses.
         furl.furl('http://[0:0:0:0:0:0:0:1:1:1:1:1:1:1:1:9999999999999]/')
 
-        # Malformed IPv6 should raise an exception because
-        # urlparse.urlsplit() raises an exception in Python v2.7+. In
-        # Python <= 2.6, urlsplit() doesn't raise a ValueError on
-        # malformed IPv6 addresses.
-        if PYTHON_27PLUS:
-            with self.assertRaises(ValueError):
-                furl.furl('http://[0:0:0:0:0:0:0:1/')
-            with self.assertRaises(ValueError):
-                furl.furl('http://0:0:0:0:0:0:0:1]/')
+        # Malformed IPv6 should raise an exception because urlparse.urlsplit()
+        # raises an exception on malformed IPv6 addresses.
+        with self.assertRaises(ValueError):
+            furl.furl('http://[0:0:0:0:0:0:0:1/')
+        with self.assertRaises(ValueError):
+            furl.furl('http://0:0:0:0:0:0:0:1]/')
 
         # Invalid host strings should raise ValueError.
         invalid_hosts = ['.', '..', 'a..b', '.a.b', '.a.b.', '$', 'a$b']
@@ -1456,12 +1709,11 @@ class TestFurl(unittest.TestCase):
         assert f.port == 888
 
         # Malformed IPv6 should raise an exception because
-        # urlparse.urlsplit() raises an exception in Python v2.7+.
-        if PYTHON_27PLUS:
-            with self.assertRaises(ValueError):
-                f.netloc = '[0:0:0:0:0:0:0:1'
-            with self.assertRaises(ValueError):
-                f.netloc = '0:0:0:0:0:0:0:1]'
+        # urlparse.urlsplit() raises an exception
+        with self.assertRaises(ValueError):
+            f.netloc = '[0:0:0:0:0:0:0:1'
+        with self.assertRaises(ValueError):
+            f.netloc = '0:0:0:0:0:0:0:1]'
 
         # Invalid ports should raise an exception.
         with self.assertRaises(ValueError):
@@ -1473,6 +1725,10 @@ class TestFurl(unittest.TestCase):
         assert f.host == '[0:0:0:0:0:0:0:1:1:1:1:1:1:1:1:9999999999999]'
         assert f.port == 888
 
+        # Empty netloc.
+        f = furl.furl('//')
+        assert f.scheme is None and f.netloc == '' and f.url == '//'
+
     def test_origin(self):
         assert furl.furl().origin == '://'
         assert furl.furl().set(host='slurp.ru').origin == '://slurp.ru'
@@ -1489,13 +1745,12 @@ class TestFurl(unittest.TestCase):
         f.origin = 'ssh://horse-machine.de'
         assert f.url == 'ssh://user:pass@horse-machine.de'
 
-        # Malformed IPv6 should raise an exception because
-        # urlparse.urlsplit() raises an exception in Python v2.7+.
-        if PYTHON_27PLUS:
-            with self.assertRaises(ValueError):
-                f.origin = '[0:0:0:0:0:0:0:1'
-            with self.assertRaises(ValueError):
-                f.origin = 'http://0:0:0:0:0:0:0:1]'
+        # Malformed IPv6 should raise an exception because urlparse.urlsplit()
+        # raises an exception.
+        with self.assertRaises(ValueError):
+            f.origin = '[0:0:0:0:0:0:0:1'
+        with self.assertRaises(ValueError):
+            f.origin = 'http://0:0:0:0:0:0:0:1]'
 
         # Invalid ports should raise an exception.
         with self.assertRaises(ValueError):
@@ -1551,7 +1806,7 @@ class TestFurl(unittest.TestCase):
         assert furl.furl('unknown://pump.com:99').set(scheme='http').port == 99
         assert furl.furl('http://pump.com:99').set(scheme='unknown').port == 99
 
-        # Domains are always lowercase.
+        # Hostnames are always lowercase.
         f = furl.furl('http://wWw.PuMpS.com:9999')
         assert f.netloc == 'www.pumps.com:9999'
         f.netloc = 'yEp.NoPe:9999'
@@ -1567,7 +1822,7 @@ class TestFurl(unittest.TestCase):
         assert self._param(f.url, 'a', 'a')
         assert self._param(f.url, 'm', 'm&m')
         assert str(f.fragment) == '1?f=frp'
-        assert str(f.path) == urllib.parse.urlsplit(f.url).path == '/sp%20ace'
+        assert str(f.path) == urlsplit(f.url).path == '/sp%20ace'
 
         assert f is f.add(path='dir', fragment_path='23', args={'b': 'b'},
                           fragment_args={'b': 'bewp'})
@@ -1719,6 +1974,14 @@ class TestFurl(unittest.TestCase):
                              username=True, password=True, port=True)
         assert f.url == 'http://host/a/big/?s=s+s'
 
+        # scheme, host, port, netloc, origin.
+        f = furl.furl('https://host:999/path')
+        assert f.copy().remove(scheme=True).url == '//host:999/path'
+        assert f.copy().remove(host=True).url == 'https://:999/path'
+        assert f.copy().remove(port=True).url == 'https://host/path'
+        assert f.copy().remove(netloc=True).url == 'https:///path'
+        assert f.copy().remove(origin=True).url == '/path'
+
         # No errors are thrown when removing URL components that don't exist.
         f = furl.furl(url)
         assert f is f.remove(fragment_path=['asdf'], fragment_args=['asdf'],
@@ -1799,15 +2062,37 @@ class TestFurl(unittest.TestCase):
                '#uwantpump?')
         assert f.url == 'http://pepp.ru/a/b/c#uwantpump?'
 
+        # In edge cases (e.g. URLs without an authority/netloc), behave
+        # identically to urllib.parse.urljoin().
+        f = furl.furl('wss://slrp.com/').join('foo:1')
+        assert f.url == 'wss://slrp.com/foo:1'
+        f = furl.furl('wss://slrp.com/').join('foo:1:rip')
+        assert f.url == 'foo:1:rip'
+        f = furl.furl('scheme:path').join('foo:blah')
+        assert f.url == 'foo:blah'
+
     def test_tostr(self):
-        f = furl.furl('http://blast.off/?a+b=c+d&two%20tap=cat%20nap%24')
+        f = furl.furl('http://blast.off/?a+b=c+d&two%20tap=cat%20nap%24%21')
         assert f.tostr() == f.url
         assert (f.tostr(query_delimiter=';') ==
-                'http://blast.off/?a+b=c+d;two+tap=cat+nap$')
+                'http://blast.off/?a+b=c+d;two+tap=cat+nap%24%21')
         assert (f.tostr(query_quote_plus=False) ==
-                'http://blast.off/?a%20b=c%20d&two%20tap=cat%20nap$')
+                'http://blast.off/?a%20b=c%20d&two%20tap=cat%20nap%24%21')
         assert (f.tostr(query_delimiter=';', query_quote_plus=False) ==
-                'http://blast.off/?a%20b=c%20d;two%20tap=cat%20nap$')
+                'http://blast.off/?a%20b=c%20d;two%20tap=cat%20nap%24%21')
+        assert (f.tostr(query_quote_plus=False, query_dont_quote=True) ==
+                'http://blast.off/?a%20b=c%20d&two%20tap=cat%20nap$!')
+        # query_dont_quote ignores invalid query characters, like '$'.
+        assert (f.tostr(query_quote_plus=False, query_dont_quote='$') ==
+                'http://blast.off/?a%20b=c%20d&two%20tap=cat%20nap$%21')
+
+        url = 'https://klugg.com/?hi=*'
+        url_encoded = 'https://klugg.com/?hi=%2A&url='
+        f = furl.furl(url).set(args=[('hi', '*'), ('url', url)])
+        assert f.tostr() == url_encoded + quote_plus(url)
+        assert f.tostr(query_dont_quote=True) == url + '&url=' + url
+        assert f.tostr(query_dont_quote='*') == (
+            url + '&url=' + quote_plus(url, '*'))
 
     def test_equality(self):
         assert furl.furl() is not furl.furl() and furl.furl() == furl.furl()
@@ -1824,34 +2109,34 @@ class TestFurl(unittest.TestCase):
         # treated as a path.
         urls = ['sup', '127.0.0.1', 'www.google.com', '192.168.1.1:8000']
         for url in urls:
-            assert isinstance(furl.urlsplit(url), urllib.parse.SplitResult)
-            assert furl.urlsplit(url).path == urllib.parse.urlsplit(url).path
+            assert isinstance(furl.urlsplit(url), SplitResult)
+            assert furl.urlsplit(url).path == urlsplit(url).path
 
         # No changes to existing urlsplit() behavior for known schemes.
         url = 'http://www.pumps.com/'
-        assert isinstance(furl.urlsplit(url), urllib.parse.SplitResult)
-        assert furl.urlsplit(url) == urllib.parse.urlsplit(url)
+        assert isinstance(furl.urlsplit(url), SplitResult)
+        assert furl.urlsplit(url) == urlsplit(url)
 
         url = 'https://www.yahoo.co.uk/one/two/three?a=a&b=b&m=m%26m#fragment'
-        assert isinstance(furl.urlsplit(url), urllib.parse.SplitResult)
-        assert furl.urlsplit(url) == urllib.parse.urlsplit(url)
+        assert isinstance(furl.urlsplit(url), SplitResult)
+        assert furl.urlsplit(url) == urlsplit(url)
 
         # Properly split the query from the path for unknown schemes.
         url = 'unknown://www.yahoo.com?one=two&three=four'
         correct = ('unknown', 'www.yahoo.com', '', 'one=two&three=four', '')
-        assert isinstance(furl.urlsplit(url), urllib.parse.SplitResult)
+        assert isinstance(furl.urlsplit(url), SplitResult)
         assert furl.urlsplit(url) == correct
 
         url = 'sup://192.168.1.102:8080///one//two////?s=kwl%20string#frag'
         correct = ('sup', '192.168.1.102:8080', '///one//two////',
                    's=kwl%20string', 'frag')
-        assert isinstance(furl.urlsplit(url), urllib.parse.SplitResult)
+        assert isinstance(furl.urlsplit(url), SplitResult)
         assert furl.urlsplit(url) == correct
 
         url = 'crazyyy://www.yahoo.co.uk/one/two/three?a=a&b=b&m=m%26m#frag'
         correct = ('crazyyy', 'www.yahoo.co.uk', '/one/two/three',
                    'a=a&b=b&m=m%26m', 'frag')
-        assert isinstance(furl.urlsplit(url), urllib.parse.SplitResult)
+        assert isinstance(furl.urlsplit(url), SplitResult)
         assert furl.urlsplit(url) == correct
 
     def test_join_path_segments(self):
@@ -2062,3 +2347,51 @@ class TestFurl(unittest.TestCase):
             assert furl.is_valid_encoded_query_value(valid)
         for invalid in invalids:
             assert not furl.is_valid_encoded_query_value(invalid)
+
+    def test_asdict(self):
+        path_encoded = '/wiki/%E3%83%AD%E3%83%AA%E3%83%9D%E3%83%83%E3%83%97'
+
+        key_encoded = '%E3%83%AD%E3%83%AA%E3%83%9D%E3%83%83%E3%83%97'
+        value_encoded = 'test%C3%A4'
+        query_encoded = 'a=1&' + key_encoded + '=' + value_encoded
+
+        host = u'ドメイン.テスト'
+        host_encoded = 'xn--eckwd4c7c.xn--zckzah'
+
+        fragment_encoded = path_encoded + '?' + query_encoded
+        url = ('https://user:pass@%s%s?%s#%s' % (
+            host_encoded, path_encoded, query_encoded, fragment_encoded))
+
+        p = furl.Path(path_encoded)
+        q = furl.Query(query_encoded)
+        f = furl.Fragment(fragment_encoded)
+        u = furl.furl(url)
+        d = {
+            'url': url,
+            'scheme': 'https',
+            'username': 'user',
+            'password': 'pass',
+            'host': host,
+            'host_encoded': host_encoded,
+            'port': 443,
+            'netloc': 'user:pass@xn--eckwd4c7c.xn--zckzah',
+            'origin': 'https://xn--eckwd4c7c.xn--zckzah',
+            'path': p.asdict(),
+            'query': q.asdict(),
+            'fragment': f.asdict(),
+            }
+        assert u.asdict() == d
+
+
+class TestMeta(unittest.TestCase):
+    def test_metadata_varibles(self):
+        def is_non_empty_string(s):
+            return isinstance(s, string_types) and s
+
+        assert is_non_empty_string(furl.__title__)
+        assert is_non_empty_string(furl.__version__)
+        assert is_non_empty_string(furl.__license__)
+        assert is_non_empty_string(furl.__author__)
+        assert is_non_empty_string(furl.__contact__)
+        assert is_non_empty_string(furl.__url__)
+        assert is_non_empty_string(furl.__description__)
